import datetime
import functools
import json
import math
import random
import re
import base64
import uuid

from pyjexl.jexl import JEXL


class ExtendedGrammar:
    def __init__(self, jexl: JEXL):
        self.jexl = jexl

    """String functions"""

    @staticmethod
    def to_string(value, prettify=False):
        if isinstance(value, (dict, list)):
            value = (
                json.dumps(value)
                if prettify
                else json.dumps(value, separators=(",", ":"))
            )
        return str(value)

    @staticmethod
    def to_json(value):
        return json.loads(value)

    @staticmethod
    def length(value):
        return len(value)

    @staticmethod
    def substring(value: any, start: int, length: int = None):
        if not isinstance(value, str):
            value = ExtendedGrammar.to_string(value)
        fin = start + length if length else len(value)
        return value[start:fin]

    @staticmethod
    def substring_before(value: any, chars: any):
        if not isinstance(value, str):
            value = ExtendedGrammar.to_string(value)
        if not isinstance(chars, str):
            chars = ExtendedGrammar.to_string(chars)
        index = value.find(chars)
        if index == -1:
            return value
        return value[:index]

    @staticmethod
    def substring_after(value: any, chars: any):
        if not isinstance(value, str):
            value = ExtendedGrammar.to_string(value)
        if not isinstance(chars, str):
            chars = ExtendedGrammar.to_string(chars)
        index = value.find(chars)
        if index == -1:
            return ""
        ini = index + len(chars)
        return value[ini:]

    @staticmethod
    def uppercase(value):
        return ExtendedGrammar.to_string(value).upper()

    @staticmethod
    def lowercase(value):
        return ExtendedGrammar.to_string(value).lower()

    @staticmethod
    def camel_case(value):
        value = ExtendedGrammar.to_string(value)
        value = re.sub(
            r"(?<!^)(?=[A-Z])|[`~!@#%^&*()|+\\\-=?;:'.,\s_']+", "_", value
        ).lower()
        parts = value.split("_")
        camel_case_value = parts[0] + "".join(x.title() for x in parts[1:])
        return camel_case_value

    @staticmethod
    def pascal_case(value):
        value = ExtendedGrammar.to_string(value)
        value = re.sub(
            r"(?<!^)(?=[A-Z])|[`~!@#%^&*()|+\\\-=?;:'.,\s_']+", "_", value
        ).lower()
        parts = value.split("_")
        camel_case_value = "".join(x.title() for x in parts)
        return camel_case_value

    @staticmethod
    def trim(value, trim_char=" "):
        return ExtendedGrammar.to_string(value).strip(trim_char)

    @staticmethod
    def pad(value, width, char=" "):
        value = ExtendedGrammar.to_string(value)
        if not isinstance(char, str):
            char = str(char)
        if width > 0:
            return value.ljust(width, char)
        else:
            return value.rjust(-width, char)

    @staticmethod
    def contains(value, search):
        return search in value

    @staticmethod
    def starts_with(value, search):
        value = ExtendedGrammar.to_string(value)
        return value.startswith(search)

    @staticmethod
    def ends_with(value, search):
        value = ExtendedGrammar.to_string(value)
        return value.endswith(search)

    @staticmethod
    def split(value: str, sep=","):
        return value.split(sep)

    @staticmethod
    def join(value, sep=","):
        return sep.join(value)

    @staticmethod
    def replace(value: str, search: str, replace=""):
        return value.replace(search, replace)

    @staticmethod
    def base64_encode(input: str):
        return base64.b64encode(input.encode("utf-8")).decode("utf-8")

    @staticmethod
    def base64_decode(input: str):
        return base64.b64decode(input.encode("utf-8")).decode("utf-8")

    @staticmethod
    def form_url_encoded(value):
        # Only works for dicts
        if isinstance(value, dict):
            return "&".join(f"{k}={v}" for k, v in value.items())
        return str(value)

    """Number functions"""

    @staticmethod
    def to_number(value):
        return float(value)

    @staticmethod
    def to_int(value):
        if isinstance(value, str):
            value = value.strip('"')
        return int(float(value))

    @staticmethod
    def abs(value):
        return abs(value)

    @staticmethod
    def floor(value):
        return math.floor(value)

    @staticmethod
    def ceil(value):
        return math.ceil(value)

    @staticmethod
    def round(value, precision=0):
        return round(value, precision)

    @staticmethod
    def power(value, power=2):
        return math.pow(value, power)

    @staticmethod
    def sqrt(value):
        return math.sqrt(value)

    @staticmethod
    def random():
        return random.random()

    @staticmethod
    def format_number(value, format="0,0.000"):
        # Determine if we need to include commas
        if "," in format:
            format = format.replace(",", "")
            formatted_value = "{:,.{precision}f}".format(
                value, precision=len(format.split(".")[1])
            )
        else:
            formatted_value = "{:.{precision}f}".format(
                value, precision=len(format.split(".")[1])
            )

        return formatted_value

    @staticmethod
    def format_base(value, base=10):
        if base == 10:
            return str(value)
        elif base == 16:
            return hex(value)[2:]  # Remove the '0x' prefix
        elif base == 8:
            return oct(value)[2:]  # Remove the '0o' prefix
        elif base == 2:
            return bin(value)[2:]  # Remove the '0b' prefix
        else:
            # Custom implementation for other bases
            digits = "0123456789abcdefghijklmnopqrstuvwxyz"
            if base > len(digits):
                raise ValueError("Base too large")
            result = ""
            while value > 0:
                result = digits[value % base] + result
                value //= base
            return result or "0"

    @staticmethod
    def format_integer(value, format="0000000"):
        # Convert the value to an integer
        integer_value = int(float(value))

        # Format the integer value according to the specified format
        formatted_value = f"{integer_value:0{len(format)}d}"

        return formatted_value

    @staticmethod
    def sum(value, *rest):
        if not isinstance(value, list):
            value = [value]
        rest = [v for v in rest]
        if len(rest) > 0 and isinstance(rest[0], list):
            rest = [v for va in rest for v in va]
        values = value + rest
        return sum(values)

    @staticmethod
    def max(value, *rest):
        if not isinstance(value, list):
            value = [value]
        rest = [v for v in rest]
        if len(rest) > 0 and isinstance(rest[0], list):
            rest = [v for va in rest for v in va]
        values = value + rest
        return max(values)

    @staticmethod
    def min(value, *rest):
        if not isinstance(value, list):
            value = [value]
        rest = [v for v in rest]
        if len(rest) > 0 and isinstance(rest[0], list):
            rest = [v for va in rest for v in va]
        values = value + rest
        return min(values)

    @staticmethod
    def avg(value, *rest):
        if not isinstance(value, list):
            value = [value]
        rest = [v for v in rest]
        if len(rest) > 0 and isinstance(rest[0], list):
            rest = [v for va in rest for v in va]
        values = value + rest
        return sum(values) / len(values)

    @staticmethod
    def to_boolean(value):
        if isinstance(value, bool):
            return value
        if isinstance(value, int):
            return value != 0
        if isinstance(value, str):
            value = value.strip().lower()
            if value == "true" or value == "1":
                return True
            if value == "false" or value == "0":
                return False
            return None
        return bool(value)

    @staticmethod
    def not_(value):
        return not ExtendedGrammar.to_boolean(value)

    """ Array functions """

    @staticmethod
    def array_range(value, start, end=None):
        if not isinstance(value, list):
            return []
        return value[start:end]

    @staticmethod
    def array_append(value, *rest):
        if not isinstance(value, list):
            value = [value]
        rest = [v for v in rest]
        if len(rest) > 0 and isinstance(rest[0], list):
            rest = [v for va in rest for v in va]
        return value + rest

    @staticmethod
    def array_reverse(value, *rest):
        if not isinstance(value, list):
            value = [value]
        rest = [v for v in rest]
        if len(rest) > 0 and isinstance(rest[0], list):
            rest = [v for va in rest for v in va]
        return (value + rest)[::-1]

    @staticmethod
    def array_shuffle(value):
        if not isinstance(value, list):
            value = [value]
        random.shuffle(value)
        return value

    @staticmethod
    def array_sort(value, reverse=False):
        if not isinstance(value, list):
            value = [value]
        return sorted(value, reverse=reverse)

    @staticmethod
    def array_distinct(value):
        if not isinstance(value, list):
            value = [value]
        return list(set(value))

    @staticmethod
    def array_to_object(input, val=None):
        if isinstance(input, str):
            return {input: val}
        if not isinstance(input, list):
            return {}
        return functools.reduce(
            lambda acc, kv: (
                acc.update({kv[0]: kv[1]}) or acc
                if isinstance(kv, list) and len(kv) == 2
                else acc.update({str(kv): val}) or acc
            ),
            input,
            {},
        )

    @staticmethod
    def array_mapfield(input, field):
        if not isinstance(input, list):
            return []
        return [item[field] for item in input]

    def array_map(self, input, expression):
        if not isinstance(input, list):
            return None
        expr = self.jexl.parse(expression)
        return [
            expr.eval({"value": value, "index": index, "array": input})
            for index, value in enumerate(input)
        ]

    def array_any(self, input, expression):
        if not isinstance(input, list):
            return False
        expr = self.jexl.parse(expression)
        return any(
            [
                expr.eval({"value": value, "index": index, "array": input})
                for index, value in enumerate(input)
            ]
        )

    def array_every(self, input, expression):
        if not isinstance(input, list):
            return False
        expr = self.jexl.parse(expression)
        return all(
            [
                expr.eval({"value": value, "index": index, "array": input})
                for index, value in enumerate(input)
            ]
        )

    def array_filter(self, input, expression):
        if not isinstance(input, list):
            return []
        expr = self.jexl.parse(expression)
        return [
            value
            for index, value in enumerate(input)
            if expr.eval({"value": value, "index": index, "array": input})
        ]

    def array_find(self, input, expression):
        if not isinstance(input, list):
            return None
        expr = self.jexl.parse(expression)
        return next(
            (
                value
                for index, value in enumerate(input)
                if expr.eval({"value": value, "index": index, "array": input})
            ),
            None,
        )

    def array_reduce(self, input, expression, initialValue=None):
        if not isinstance(input, list):
            return None
        expr = self.jexl.parse(expression)
        return functools.reduce(
            lambda acc, value: expr.eval({"accumulator": acc, "value": value}),
            input,
            initialValue,
        )

    """ Object functions """

    @staticmethod
    def object_keys(input):
        if isinstance(input, dict):
            return list(input.keys())
        return None

    @staticmethod
    def object_values(input):
        if isinstance(input, dict):
            return list(input.values())
        return None

    @staticmethod
    def object_entries(input):
        if isinstance(input, dict):
            return list(input.items())
        return None

    @staticmethod
    def object_merge(*args):
        result = {}
        for arg in args:
            if isinstance(arg, list):
                for obj in arg:
                    if isinstance(obj, dict):
                        result.update(obj)
            elif isinstance(arg, dict):
                result.update(arg)
        return result

    """ Date functions """

    @staticmethod
    def now():
        return datetime.datetime.isoformat(datetime.datetime.now())

    @staticmethod
    def millis():
        return datetime.datetime.now().timestamp() * 1000

    @staticmethod
    def to_datetime(value):
        return datetime.datetime.fromtimestamp(value / 1000).isoformat()

    @staticmethod
    def to_millis(value):
        return datetime.datetime.fromisoformat(value).timestamp() * 1000

    @staticmethod
    def datetime_add(input, unit, value):
        input_datetime = datetime.datetime.fromisoformat(input)
        if not str.endswith(unit, "s"):
            unit = unit + "s"
        if unit == "years":
            return input_datetime + datetime.timedelta(days=365 * value)
        if unit == "months":
            return input_datetime + datetime.timedelta(days=30 * value)
        if unit == "days":
            return input_datetime + datetime.timedelta(days=value)
        if unit == "hours":
            return input_datetime + datetime.timedelta(hours=value)
        if unit == "minutes":
            return input_datetime + datetime.timedelta(minutes=value)
        if unit == "seconds":
            return input_datetime + datetime.timedelta(seconds=value)
        if unit == "milliseconds":
            return input_datetime + datetime.timedelta(milliseconds=value)
        return None

    @staticmethod
    def convert_time_zone(input, target_timezone):
        """
        Converts an ISO datetime string to a target timezone, handling daylight savings, and returns an ISO string with the correct offset.
        """
        try:
            from dateutil import parser as dtparser
            import pytz
        except ImportError:
            raise ImportError("dateutil and pytz are required for timezone conversion.")

        # Minimal Windows to IANA mapping (extend as needed)
        WINDOWS_TZ_MAP = {
            "Pacific Standard Time": "America/Los_Angeles",
            "UTC": "UTC",
            # Add more mappings as needed
        }

        # Parse input datetime
        if not isinstance(input, str):
            input = ExtendedGrammar.to_string(input)
        try:
            dt = dtparser.isoparse(input)
        except Exception:
            return None

        # Ensure datetime is aware (UTC if Z or no offset)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=pytz.UTC)

        tz_str = ExtendedGrammar.to_string(target_timezone)

        # Check for fixed offset (e.g., +02:00, -08:00)
        offset_match = re.match(r"^([+-])(\d{2}):(\d{2})$", tz_str)
        if offset_match:
            sign = 1 if offset_match.group(1) == "+" else -1
            hours = int(offset_match.group(2))
            minutes = int(offset_match.group(3))
            offset = datetime.timedelta(hours=sign * hours, minutes=sign * minutes)
            tzinfo = datetime.timezone(offset)
            dt_converted = dt.astimezone(tzinfo)
        else:
            # Try Windows mapping
            iana_tz = WINDOWS_TZ_MAP.get(tz_str, tz_str)
            try:
                tzinfo = pytz.timezone(iana_tz)
            except Exception:
                # Try to find by case-insensitive match
                try:
                    tzinfo = next(
                        z
                        for z in map(pytz.timezone, pytz.all_timezones)
                        if z.zone.lower() == iana_tz.lower()
                    )
                except Exception:
                    return None
            dt_converted = dt.astimezone(tzinfo)

        # Format: ISO string with 7 digits microseconds and offset
        iso_str = dt_converted.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
        # Insert colon in offset for ISO compliance
        if iso_str.endswith("+0000"):
            iso_str = iso_str[:-5] + "+00:00"
        elif iso_str[-5] in ["+", "-"]:
            iso_str = iso_str[:-5] + iso_str[-5:-2] + ":" + iso_str[-2:]
        # Truncate microseconds to 7 digits
        if "." in iso_str:
            ms = iso_str.split(".")[1][:7]
            iso_str = iso_str.split(".")[0] + "." + ms + iso_str[-6:]
        return iso_str

    """ Misc """

    def _eval(self, input, expression):
        if not isinstance(expression, str) and isinstance(input, str):
            return self.jexl.evaluate(expression)
        if isinstance(input, dict) and isinstance(expression, str):
            return self.jexl.evaluate(expression, input)
        return None

    def uuid(self):
        return str(uuid.uuid4())



import operator


class Operator(object):
    __slots__ = ('symbol', 'precedence', 'evaluate', '_evaluate_lazy')

    def __init__(self, symbol, precedence, evaluate, evaluate_lazy=False):
        """Operator definition.

        If evaluate_lazy is set to True, the `evaluate()` method will receive
        it's parameters as a lambda expression that needs to be called to
        receive the value of the expression. Otherwise, the values will
        already be evaluated.
        """
        self.symbol = symbol
        self.precedence = precedence
        self.evaluate = evaluate
        self._evaluate_lazy = evaluate_lazy

    def do_evaluate(self, *args):
        if self._evaluate_lazy:
            return self.evaluate(*args)
        return self.evaluate(*(arg() for arg in args))

    def __repr__(self):
        return 'Operator({})'.format(repr(self.symbol))


default_binary_operators = {
    '+': Operator('+', 30, operator.add),
    '-': Operator('-', 30, operator.sub),
    '*': Operator('*', 40, operator.mul),
    '//': Operator('//', 40, operator.floordiv),
    '/': Operator('/', 40, operator.truediv),
    '%': Operator('%', 50, operator.mod),
    '^': Operator('^', 50, operator.pow),
    '==': Operator('==', 20, operator.eq),
    '!=': Operator('!=', 20, operator.ne),
    '>=': Operator('>=', 20, operator.ge),
    '>': Operator('>', 20, operator.gt),
    '<=': Operator('<=', 20, operator.le),
    '<': Operator('<', 20, operator.lt),
    '&&': Operator('&&', 10, lambda a, b: a() and b(), evaluate_lazy=True),
    '||': Operator('||', 10, lambda a, b: a() or b(), evaluate_lazy=True),
    'in': Operator('in', 20, lambda a, b: a in b),
}


default_unary_operators = {
    '!': Operator('!', 1000, operator.not_),
}



from builtins import str

import pytest

from pyjexl.evaluator import Context, Evaluator
from pyjexl.exceptions import MissingTransformError
from pyjexl.jexl import JEXLConfig
from pyjexl.operators import default_binary_operators, default_unary_operators

from . import DefaultEvaluator, DefaultParser


def tree(expression):
    return DefaultParser().parse(expression)


def test_literal():
    result = DefaultEvaluator().evaluate(tree("1"))
    assert result == 1.0


def test_binary_expression():
    result = DefaultEvaluator().evaluate(tree("1 + 2"))
    assert result == 3.0


def test_arithmetic():
    result = DefaultEvaluator().evaluate(tree("(2 + 3) * 4"))
    assert result == 20


def test_string_concat():
    # Because we don't have implicit type conversions like JavaScript,
    # we diverge from the original JEXL test suite and add a filter.
    config = JEXLConfig(
        {}, {"str": str}, default_binary_operators, default_unary_operators
    )
    evaluator = Evaluator(config)
    result = evaluator.evaluate(tree('"Hello" + (4+4)|str + "Wo\\"rld"'))
    assert result == 'Hello8Wo"rld'


def test_true_comparison():
    result = DefaultEvaluator().evaluate(tree("2 > 1"))
    assert result


def test_false_comparison():
    result = DefaultEvaluator().evaluate(tree("2 <= 1"))
    assert not result


def test_complex():
    result = DefaultEvaluator().evaluate(tree('"foo" && 6 >= 6 && 0 + 1 && true'))
    assert result


def test_identifier_chain():
    context = Context({"foo": {"baz": {"bar": "tek"}}})
    result = DefaultEvaluator().evaluate(tree("foo.baz.bar"), context)
    assert result == "tek"


def test_transforms():
    config = JEXLConfig(
        {}, {"half": lambda x: x / 2}, default_binary_operators, default_unary_operators
    )
    evaluator = Evaluator(config)
    result = evaluator.evaluate(tree("foo|half + 3"), {"foo": 10})
    assert result == 8


def test_functions():
    config = JEXLConfig(
        {"half": lambda x: x / 2}, {}, default_binary_operators, default_unary_operators
    )
    evaluator = Evaluator(config)
    result = evaluator.evaluate(tree("half(foo) + 3"), {"foo": 10})
    assert result == 8


def test_functions2():
    config = JEXLConfig(
        {"two": lambda: 2}, {}, default_binary_operators, default_unary_operators
    )
    evaluator = Evaluator(config)
    result = evaluator.evaluate(tree("two() + 3"))
    assert result == 5


def test_filter_arrays():
    context = Context(
        {"foo": {"bar": [{"tek": "hello"}, {"tek": "baz"}, {"tok": "baz"}]}}
    )

    result = DefaultEvaluator().evaluate(tree('foo.bar[.tek == "baz"]'), context)
    assert result == [{"tek": "baz"}]


def test_array_index():
    context = Context(
        {"foo": {"bar": [{"tek": "tok"}, {"tek": "baz"}, {"tek": "foz"}]}}
    )

    result = DefaultEvaluator().evaluate(tree("foo.bar[1].tek"), context)
    assert result == "baz"


def test_filter_object_properties():
    context = Context({"foo": {"baz": {"bar": "tek"}}})
    result = DefaultEvaluator().evaluate(tree('foo["ba" + "z"].bar'), context)
    assert result == "tek"


def test_missing_transform_exception():
    with pytest.raises(MissingTransformError):
        DefaultEvaluator().evaluate(tree('"hello"|world'))


def test_divfloor():
    result = DefaultEvaluator().evaluate(tree("7 // 2"))
    assert result == 3


def test_object_literal():
    result = DefaultEvaluator().evaluate(tree('{foo: {bar: "tek"}}'))
    assert result == {"foo": {"bar": "tek"}}


def test_empty_object_literal():
    result = DefaultEvaluator().evaluate(tree("{}"))
    assert result == {}


def test_transforms_multiple_arguments():
    config = JEXLConfig(
        binary_operators=default_binary_operators,
        unary_operators=default_unary_operators,
        transforms={
            "concat": lambda val, a1, a2, a3: val + ": " + a1 + a2 + a3,
        },
        functions={},
    )
    evaluator = Evaluator(config)
    result = evaluator.evaluate(tree('"foo"|concat("baz", "bar", "tek")'))
    assert result == "foo: bazbartek"


def test_functions_multiple_arguments():
    config = JEXLConfig(
        binary_operators=default_binary_operators,
        unary_operators=default_unary_operators,
        transforms={},
        functions={
            "concat": lambda val, a1, a2, a3: val + ": " + a1 + a2 + a3,
        },
    )
    evaluator = Evaluator(config)
    result = evaluator.evaluate(tree('concat("foo", "baz", "bar", "tek")'))
    assert result == "foo: bazbartek"


def test_object_literal_properties():
    result = DefaultEvaluator().evaluate(tree('{foo: "bar"}.foo'))
    assert result == "bar"


def test_array_literal():
    result = DefaultEvaluator().evaluate(tree('["foo", 1+2]'))
    assert result == ["foo", 3]


def test_in_operator_string():
    result = DefaultEvaluator().evaluate(tree('"bar" in "foobartek"'))
    assert result is True

    result = DefaultEvaluator().evaluate(tree('"baz" in "foobartek"'))
    assert result is False


def test_in_operator_array():
    result = DefaultEvaluator().evaluate(tree('"bar" in ["foo","bar","tek"]'))
    assert result is True

    result = DefaultEvaluator().evaluate(tree('"baz" in ["foo","bar","tek"]'))
    assert result is False


def test_conditional_expression():
    result = DefaultEvaluator().evaluate(tree('"foo" ? 1 : 2'))
    assert result == 1

    result = DefaultEvaluator().evaluate(tree('"" ? 1 : 2'))
    assert result == 2


def test_arbitrary_whitespace():
    result = DefaultEvaluator().evaluate(tree("(\t2\n+\n3) *\n4\n\r\n"))
    assert result == 20


@pytest.mark.parametrize(
    "expression,expect_result,expect_fail",
    [
        ("true || 1/0", True, False),
        ("true && 1/0", None, True),
        ("false || 1/0", None, True),
        ("false && 1/0", False, False),
    ],
)
def test_logic_shortcuts(expression, expect_result, expect_fail):
    if expect_fail:
        with pytest.raises(Exception):
            DefaultEvaluator().evaluate(tree(expression))
    else:
        assert expect_result == DefaultEvaluator().evaluate(tree(expression))



import pytest
import hypothesis

from pyjexl.analysis import JEXLAnalyzer
from pyjexl.exceptions import MissingTransformError, ParseError
from pyjexl.jexl import JEXL


@pytest.fixture(scope="module")
def jexl():
    """Return an instance of the JEXL class. Useful when tests are
    repeated with hypothesis since only one instance of JEXL is created
    once."""
    return JEXL()


def test_it_works(jexl):
    assert JEXL().evaluate('1 + 1') == 2


def test_parse_error():
    jexl = JEXL()
    with pytest.raises(ParseError):
        jexl.evaluate('this is ( invalid + =s')


def test_add_transform():
    jexl = JEXL()
    jexl.add_transform('foo', lambda x: x + 2)
    assert jexl.evaluate('4|foo') == 6


def test_remove_transform():
    jexl = JEXL()
    jexl.add_transform('foo', lambda x: x)
    assert jexl.evaluate('4|foo') == 4

    jexl.remove_transform('foo')
    with pytest.raises(MissingTransformError):
        jexl.evaluate('4|foo')


def test_transform_decorator_explicit_name():
    jexl = JEXL()

    @jexl.transform(name='new_name')
    def old_name(x):
        return x + 2

    assert jexl.evaluate('3|new_name') == 5
    with pytest.raises(MissingTransformError):
        jexl.evaluate('3|old_name')


def test_transform_decorator_using_function_name():
    jexl = JEXL()

    @jexl.transform()
    def old_name(x):
        return x + 2

    assert jexl.evaluate('3|old_name') == 5


def test_add_custom_binary_operator():
    jexl = JEXL()
    jexl.add_binary_operator('=', 20, lambda x, y: (x + y) / 2)
    assert jexl.evaluate('2 = 4') == 3


def test_remove_binary_operator():
    jexl = JEXL()
    jexl.remove_binary_operator('+')
    with pytest.raises(ParseError):
        jexl.evaluate('2 + 4')


def test_binary_operator_precedence():
    jexl = JEXL()
    jexl.add_binary_operator('=', 50, lambda x, y: x + y)
    jexl.add_binary_operator('@', 100, lambda x, y: x / y)

    assert jexl.evaluate('(3 = 6) @ 3') == 3
    assert jexl.evaluate('3 = 6 @ 3') == 5


def test_add_custom_unary_operator():
    jexl = JEXL()
    jexl.add_unary_operator('=', lambda x: x + 5)
    assert jexl.evaluate('=5') == 10


def test_remove_unary_operator():
    jexl = JEXL()
    jexl.remove_unary_operator('!')
    with pytest.raises(ParseError):
        jexl.evaluate('!true')


def test_grammar_invalidation():
    """
    Ensure that the grammar object doesn't get stale when operators
    change.
    """
    jexl = JEXL()
    jexl.add_unary_operator('=', lambda x: x + 5)
    assert jexl.evaluate('=5') == 10

    jexl.remove_unary_operator('=')
    with pytest.raises(ParseError):
        jexl.evaluate('=5')


def test_validate():
    jexl = JEXL()
    jexl.add_transform('foo', lambda x: x + 1)
    assert list(jexl.validate('5+6|foo')) == []

    errors = list(jexl.validate('5+6|bar'))
    assert len(errors) == 1
    assert 'bar' in errors[0]

    errors = list(jexl.validate('1+'))
    assert errors == ['Could not parse expression: 1+']

    errors = list(jexl.validate('"\n"'))
    assert errors == ['Could not parse expression: "\n"']


JEXL_ALPHABET = hypothesis.strategies.characters(whitelist_categories=(
    # Letters
    'Lu', 'Ll',
    # Numbers
    'Nd', 'No',
    # Spaces
    'Zs', 'Zl', 'Zp',
), whitelist_characters=[
    '\t', '\n', '\r', ' ',
    '$', '|', '+', '/',
    '(', ')', '{', '}', '[', ']',
    "'", '"', ',', ':', '?',
])


@hypothesis.given(hypothesis.strategies.text(JEXL_ALPHABET))
@hypothesis.settings(max_examples=500)
def test_validate_never_throws(jexl, s):
    errors = list(jexl.validate(s))
    assert isinstance(errors, list)

    # '{{' and '}}' are vanishingly rare sequences when generating
    # by-character the way we do. Artificially increase coverage of
    # these important sequences by explicitly generating them based on
    # single-character '{' and '}'.
    #
    # FIXME: figure out how to randomly generate a sequence of tokens
    # instead of characters.
    errors = list(jexl.validate(s.replace('{', '{{').replace('}', '}}')))
    assert isinstance(errors, list)


def test_validate_simple_equality():
    jexl = JEXL()
    errors = list(jexl.validate('FOO_BAR ==12345'))
    assert errors == []


class SumIntAnalyzer(JEXLAnalyzer):
    """Test analyzer that sums up all integers in a statement."""
    def generic_visit(self, expression):
        children = list(expression.children)
        if children:
            return sum(self.visit(child) for child in children)
        else:
            return 0

    def visit_Literal(self, literal):
        if isinstance(literal.value, int):
            return literal.value
        else:
            return 0


def test_analysis():
    jexl = JEXL()
    assert jexl.analyze('1+(2*3)|concat(4)', SumIntAnalyzer) == 10



import ast
from future.builtins.misc import super
from future.utils import with_metaclass

from parsimonious import Grammar, NodeVisitor

from pyjexl.exceptions import InvalidOperatorError
from pyjexl.operators import Operator


def operator_pattern(operators):
    """Generate a grammar rule to match a dict of operators."""
    # Sort operators and reverse them so that operators sharing a prefix
    # always have the shortest forms last. Otherwise, the shorter forms
    # will match operators early, i.e. `<` will match `<=` and ignore
    # the `=`, causing a parse error.
    operators = sorted(operators, key=lambda op: op.symbol, reverse=True)
    operator_literals = ['"{}"'.format(op.symbol) for op in operators]
    return " / ".join(operator_literals)


def jexl_grammar(jexl_config):
    return Grammar(
        r"""
        expression = (
            _ (conditional_expression / binary_expression / unary_expression / complex_value) _
        )

        conditional_expression = (
            conditional_test _ "?" _ expression _ ":" _ expression
        )
        conditional_test = (binary_expression / unary_expression / complex_value)

        binary_expression = binary_operand (_ binary_operator _ binary_operand)+
        binary_operator = {binary_op_pattern}
        binary_operand = (unary_expression / complex_value)

        unary_expression = unary_operator _ unary_operand
        unary_operator = {unary_op_pattern}
        unary_operand = (unary_expression / complex_value)

        complex_value = value (transform / attribute / filter_expression)*

        transform = "|" identifier transform_arguments?
        transform_arguments = "(" _ value_list _ ")"

        function = identifier function_arguments
        function_arguments = "(" _ value_list? _ ")"

        attribute = "." identifier

        filter_expression = "[" _ expression _ "]"

        value = (
            function / boolean / string / numeric / subexpression / object_literal /
            array_literal / identifier / relative_identifier
        )

        subexpression = "(" _ expression _ ")"

        object_literal = "{{" _ object_key_value_list? _ "}}"
        object_key_value_list = object_key_value (_ "," _ object_key_value)*
        object_key_value = identifier _ ":" _ expression

        array_literal = "[" _ value_list? _ "]"
        value_list = expression (_ "," _ expression)*

        identifier = ~r"[a-zA-Z_\$][a-zA-Z0-9_\$]*"
        relative_identifier = "." identifier

        boolean = "true" / "false"
        string = ~"\"[^\"\\\\\\n\\r]*(?:\\\\.[^\"\\\\\\n\\r]*)*\""is /
                 ~"'[^'\\\\\\n\\r]*(?:\\\\.[^'\\\\\\n\\r]*)*'"is
        numeric = "-"? number ("." number)?

        number = ~r"[0-9]+"

        _ = ~r"\s*"
    """.format(
            binary_op_pattern=operator_pattern(jexl_config.binary_operators.values()),
            unary_op_pattern=operator_pattern(jexl_config.unary_operators.values()),
        )
    )


class Parser(NodeVisitor):
    def __init__(self, jexl_config):
        self.config = jexl_config
        self._relative = 0

    def visit_expression(self, node, children):
        return children[1][0]

    def visit_subexpression(self, node, children):
        (left_paren, _, expression, _, right_paren) = children
        return expression

    def visit_binary_expression(self, node, children):
        pieces = [children[0]]
        for op_value in node.children[1].children:
            op_symbol = self.visit(op_value.children[1])
            value = self.visit(op_value.children[3])
            pieces.append(op_symbol)
            pieces.append(value)

        # Build a tree of binary operators based on precedence. Adapted
        # from JEXL's parser code for handling binary operators.
        cursor = pieces.pop(0)
        for piece in pieces:
            if not isinstance(piece, Operator):
                cursor.right = piece
                piece.parent = cursor
                cursor = piece
                continue

            # Parents are always operators
            parent = cursor.parent
            while parent is not None and parent.operator.precedence >= piece.precedence:
                cursor = parent
                parent = cursor.parent

            node = BinaryExpression(operator=piece, left=cursor, parent=parent)
            cursor.parent = node
            if parent is not None:
                parent.right = node
            cursor = node

        return cursor.root()

    def visit_conditional_expression(self, node, children):
        (test, _, question, _, consequent, _, colon, _, alternate) = children
        return ConditionalExpression(
            test=test, consequent=consequent, alternate=alternate
        )

    def visit_conditional_test(self, node, children):
        return children[0]

    def visit_binary_operator(self, node, children):
        try:
            return self.config.binary_operators[node.text]
        except KeyError:
            raise InvalidOperatorError(node.text)

    def visit_binary_operand(self, node, children):
        return children[0]

    def visit_unary_expression(self, node, children):
        (operator, _, value) = children
        return UnaryExpression(operator=operator, right=value)

    def visit_unary_operator(self, node, children):
        try:
            return self.config.unary_operators[node.text]
        except KeyError:
            raise InvalidOperatorError(node.text)

    def visit_unary_operand(self, node, children):
        return children[0]

    def visit_object_literal(self, node, children):
        (left_brace, _, value, _, right_brace) = children

        return ObjectLiteral(value=value[0] if isinstance(value, list) else {})

    def visit_object_key_value_list(self, node, children):
        key_values = [children[0]]
        for _, comma, _, key_value in children[1]:
            key_values.append(key_value)

        return {identifier.value: value for identifier, value in key_values}

    def visit_object_key_value(self, node, children):
        (identifier, _, colon, _, value) = children
        return (identifier, value)

    def visit_array_literal(self, node, children):
        (left_bracket, _, value, _, right_bracket) = children
        return ArrayLiteral(value=value[0] if isinstance(value, list) else [])

    def visit_value_list(self, node, children):
        values = [children[0]]
        for _, comma, _, value in children[1]:
            values.append(value)

        return values

    def visit_complex_value(self, node, children):
        (value, modifiers) = children

        current = value
        for (modifier,) in modifiers:
            modifier.subject = current
            current = modifier

        return current

    def visit_attribute(self, node, children):
        (dot, identifier) = children
        return identifier

    def visit_function(self, node, children):
        (identifier, arguments) = children
        func = Function(name=identifier.value, args=[])

        if isinstance(arguments, list):
            try:
                func.args = arguments
            except (IndexError, TypeError):
                pass

        return func

    def visit_function_arguments(self, node, children):
        (left_paren, _, values, _, right_paren) = children
        values = [item for sublist in values for item in sublist]
        return values

    def visit_transform(self, node, children):
        (bar, identifier, arguments) = children
        transform = Transform(name=identifier.value, args=[])

        try:
            transform.args = arguments[0]
        except (IndexError, TypeError):
            pass

        return transform

    def visit_transform_arguments(self, node, children):
        (left_paren, _, values, _, right_paren) = children
        return values

    def visit_filter_expression(self, node, children):
        (left_bracket, _, expression, _, right_bracket) = children
        return FilterExpression(
            expression=expression, relative=expression.contains_relative()
        )

    def visit_identifier(self, node, children):
        return Identifier(value=node.text, relative=False)

    def visit_relative_identifier(self, node, children):
        (dot, identifier) = children
        identifier.relative = True
        return identifier

    def visit_value(self, node, children):
        return children[0]

    def visit_boolean(self, node, children):
        if node.text == "true":
            return Literal(True)
        elif node.text == "false":
            return Literal(False)
        else:
            raise ValueError("Could not parse boolean: " + node.text)

    def visit_numeric(self, node, children):
        number_type = float if "." in node.text else int
        return Literal(number_type(node.text))

    def visit_string(self, node, children):
        return Literal(ast.literal_eval(node.text))

    def generic_visit(self, node, visited_children):
        return visited_children or node


class NodeMeta(type):
    """
    Metaclass for making AST nodes. Handles adding the Node's custom
    fields to __slots__ and ensures every Node has a parent field.
    """

    def __new__(meta, classname, bases, classdict):
        if "parent" not in classdict["fields"]:
            classdict["fields"].append("parent")
        classdict.update(
            {
                "__slots__": classdict["fields"],
            }
        )
        return type.__new__(meta, classname, bases, classdict)


class Node(with_metaclass(NodeMeta, object)):
    """
    Base class for AST Nodes.

    Nodes are like mutable namedtuples. Each node type should declare a
    fields attribute on the class that lists the desired attributes for
    the class.
    """

    fields = []

    def __init__(self, *args, **kwargs):
        """
        Accepts values for this node's fields as both positional (in the
        order defined in self.fields) and keyword arguments.
        """
        for index, field in enumerate(self.fields):
            if len(args) > index:
                setattr(self, field, args[index])
            else:
                setattr(self, field, kwargs.get(field, None))

    def __repr__(self):
        kwargs = [
            "=".join([field, repr(getattr(self, field))])
            for field in self.fields
            if field != "parent"
        ]

        return "{name}({kwargs})".format(
            name=type(self).__name__, kwargs=", ".join(kwargs)
        )

    def __eq__(self, other):
        return isinstance(other, type(self)) and all(
            getattr(self, field) == getattr(other, field)
            for field in self.fields
            if field != "parent"
        )

    @property
    def children(self):
        return iter(())

    def root(self):
        if self.parent is None:
            return self
        else:
            return self.parent.root()

    def contains_relative(self):
        children_relative = any(
            child.contains_relative() for child in self.children if child is not None
        )
        return getattr(self, "relative", children_relative)


class BinaryExpression(Node):
    fields = ["operator", "left", "right"]

    @property
    def children(self):
        yield self.left
        yield self.right


class UnaryExpression(Node):
    fields = ["operator", "right"]

    @property
    def children(self):
        yield self.right


class Literal(Node):
    fields = ["value"]


class Identifier(Node):
    fields = ["value", "subject", "relative"]

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("relative", False)
        super().__init__(*args, **kwargs)

    @property
    def children(self):
        if self.subject is not None:
            yield self.subject


class ObjectLiteral(Node):
    fields = ["value"]


class ArrayLiteral(Node):
    fields = ["value"]


class Transform(Node):
    fields = ["name", "args", "subject"]

    @property
    def children(self):
        yield self.subject
        for arg in self.args:
            yield arg


class Function(Node):
    fields = ["name", "args"]

    @property
    def children(self):
        for arg in self.args:
            yield arg


class FilterExpression(Node):
    fields = ["expression", "subject", "relative"]

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("relative", False)
        super().__init__(*args, **kwargs)

    @property
    def children(self):
        yield self.expression
        yield self.subject

    def contains_relative(self):
        return self.subject.contains_relative()


class ConditionalExpression(Node):
    fields = ["test", "consequent", "alternate"]

    @property
    def children(self):
        yield self.test
        yield self.consequent
        yield self.alternate



from .jexl import JEXL
from .extended_grammar import ExtendedGrammar
import random
import math


class JexlExtended(JEXL):
    def __init__(self, context=None):
        super().__init__(context=context)
        self._extended_grammar = ExtendedGrammar(self)

        """ String functions """

        super().add_function("string", ExtendedGrammar.to_string)
        super().add_function("$string", ExtendedGrammar.to_string)
        super().add_transform("toString", ExtendedGrammar.to_string)
        super().add_transform("string", ExtendedGrammar.to_string)
        super().add_function("json", ExtendedGrammar.to_json)
        super().add_function("$json", ExtendedGrammar.to_json)
        super().add_transform("toJson", ExtendedGrammar.to_json)
        super().add_function("parseJson", ExtendedGrammar.to_json)
        super().add_function("$parseJson", ExtendedGrammar.to_json)
        super().add_transform("parseJson", ExtendedGrammar.to_json)
        super().add_function("length", ExtendedGrammar.length)
        super().add_function("$length", ExtendedGrammar.length)
        super().add_transform("length", ExtendedGrammar.length)
        super().add_function("count", ExtendedGrammar.length)
        super().add_function("$count", ExtendedGrammar.length)
        super().add_transform("count", ExtendedGrammar.length)
        super().add_function("size", ExtendedGrammar.length)
        super().add_function("$size", ExtendedGrammar.length)
        super().add_transform("size", ExtendedGrammar.length)
        super().add_function("substring", ExtendedGrammar.substring)
        super().add_function("$substring", ExtendedGrammar.substring)
        super().add_transform("substring", ExtendedGrammar.substring)
        super().add_function("substringBefore", ExtendedGrammar.substring_before)
        super().add_function("$substringBefore", ExtendedGrammar.substring_before)
        super().add_transform("substringBefore", ExtendedGrammar.substring_before)
        super().add_function("substringAfter", ExtendedGrammar.substring_after)
        super().add_function("$substringAfter", ExtendedGrammar.substring_after)
        super().add_transform("substringAfter", ExtendedGrammar.substring_after)
        super().add_function("uppercase", ExtendedGrammar.uppercase)
        super().add_function("$uppercase", ExtendedGrammar.uppercase)
        super().add_transform("uppercase", ExtendedGrammar.uppercase)
        super().add_function("upper", ExtendedGrammar.uppercase)
        super().add_function("$upper", ExtendedGrammar.uppercase)
        super().add_transform("upper", ExtendedGrammar.uppercase)
        super().add_function("lowercase", ExtendedGrammar.lowercase)
        super().add_function("$lowercase", ExtendedGrammar.lowercase)
        super().add_transform("lowercase", ExtendedGrammar.lowercase)
        super().add_function("lower", ExtendedGrammar.lowercase)
        super().add_function("$lower", ExtendedGrammar.lowercase)
        super().add_transform("lower", ExtendedGrammar.lowercase)
        super().add_function("camelCase", ExtendedGrammar.camel_case)
        super().add_function("$camelCase", ExtendedGrammar.camel_case)
        super().add_transform("camelCase", ExtendedGrammar.camel_case)
        super().add_transform("camelcase", ExtendedGrammar.camel_case)
        super().add_transform("toCamelCase", ExtendedGrammar.camel_case)
        super().add_function("pascalCase", ExtendedGrammar.pascal_case)
        super().add_function("$pascalCase", ExtendedGrammar.pascal_case)
        super().add_transform("pascalCase", ExtendedGrammar.pascal_case)
        super().add_transform("pascalcase", ExtendedGrammar.pascal_case)
        super().add_transform("toPascalCase", ExtendedGrammar.pascal_case)
        super().add_function("trim", ExtendedGrammar.trim)
        super().add_function("$trim", ExtendedGrammar.trim)
        super().add_transform("trim", ExtendedGrammar.trim)
        super().add_function("pad", ExtendedGrammar.pad)
        super().add_function("$pad", ExtendedGrammar.pad)
        super().add_transform("pad", ExtendedGrammar.pad)
        super().add_function("contains", ExtendedGrammar.contains)
        super().add_function("$contains", ExtendedGrammar.contains)
        super().add_transform("contains", ExtendedGrammar.contains)
        # DateTime timezone conversion
        super().add_function("convertTimeZone", ExtendedGrammar.convert_time_zone)
        super().add_function("$convertTimeZone", ExtendedGrammar.convert_time_zone)
        super().add_transform("convertTimeZone", ExtendedGrammar.convert_time_zone)
        super().add_function("includes", ExtendedGrammar.contains)
        super().add_function("$includes", ExtendedGrammar.contains)
        super().add_transform("includes", ExtendedGrammar.contains)
        super().add_function("startsWith", ExtendedGrammar.starts_with)
        super().add_function("$startsWith", ExtendedGrammar.starts_with)
        super().add_transform("startsWith", ExtendedGrammar.starts_with)
        super().add_function("endsWith", ExtendedGrammar.ends_with)
        super().add_function("$endsWith", ExtendedGrammar.ends_with)
        super().add_transform("endsWith", ExtendedGrammar.ends_with)
        super().add_function("split", ExtendedGrammar.split)
        super().add_function("$split", ExtendedGrammar.split)
        super().add_transform("split", ExtendedGrammar.split)
        super().add_function("join", ExtendedGrammar.join)
        super().add_function("$join", ExtendedGrammar.join)
        super().add_transform("join", ExtendedGrammar.join)
        super().add_function("replace", ExtendedGrammar.replace)
        super().add_function("$replace", ExtendedGrammar.replace)
        super().add_transform("replace", ExtendedGrammar.replace)
        super().add_function("base64Encode", ExtendedGrammar.base64_encode)
        super().add_function("$base64Encode", ExtendedGrammar.base64_encode)
        super().add_transform("base64Encode", ExtendedGrammar.base64_encode)
        super().add_function("base64Decode", ExtendedGrammar.base64_decode)
        super().add_function("$base64Decode", ExtendedGrammar.base64_decode)
        super().add_transform("base64Decode", ExtendedGrammar.base64_decode)
        super().add_function("formUrlEncoded", ExtendedGrammar.form_url_encoded)
        super().add_function("$formUrlEncoded", ExtendedGrammar.form_url_encoded)
        super().add_transform("formUrlEncoded", ExtendedGrammar.form_url_encoded)

        """ Number functions """

        super().add_function("number", ExtendedGrammar.to_number)
        super().add_function("$number", ExtendedGrammar.to_number)
        super().add_transform("number", ExtendedGrammar.to_number)
        super().add_transform("toNumber", ExtendedGrammar.to_number)
        super().add_function("parseFloat", ExtendedGrammar.to_number)
        super().add_function("$parseFloat", ExtendedGrammar.to_number)
        super().add_transform("float", ExtendedGrammar.to_number)
        super().add_transform("toFloat", ExtendedGrammar.to_number)
        super().add_function("parseInteger", ExtendedGrammar.to_int)
        super().add_function("parseInt", ExtendedGrammar.to_int)
        super().add_function("$parseInteger", ExtendedGrammar.to_int)
        super().add_transform("parseInteger", ExtendedGrammar.to_int)
        super().add_transform("parseInt", ExtendedGrammar.to_int)
        super().add_transform("toInt", ExtendedGrammar.to_int)
        super().add_transform("integer", ExtendedGrammar.to_int)
        super().add_function("abs", ExtendedGrammar.abs)
        super().add_function("$abs", ExtendedGrammar.abs)
        super().add_transform("abs", ExtendedGrammar.abs)
        super().add_function("floor", ExtendedGrammar.floor)
        super().add_function("$floor", ExtendedGrammar.floor)
        super().add_transform("floor", ExtendedGrammar.floor)
        super().add_function("ceil", ExtendedGrammar.ceil)
        super().add_function("$ceil", ExtendedGrammar.ceil)
        super().add_transform("ceil", ExtendedGrammar.ceil)
        super().add_function("round", ExtendedGrammar.round)
        super().add_function("$round", ExtendedGrammar.round)
        super().add_transform("round", ExtendedGrammar.round)
        super().add_function("power", ExtendedGrammar.power)
        super().add_function("$power", ExtendedGrammar.power)
        super().add_transform("power", ExtendedGrammar.power)
        super().add_function("sqrt", ExtendedGrammar.sqrt)
        super().add_function("$sqrt", ExtendedGrammar.sqrt)
        super().add_transform("sqrt", ExtendedGrammar.sqrt)
        super().add_function("random", ExtendedGrammar.random)
        super().add_function("$random", ExtendedGrammar.random)
        super().add_function("formatNumber", ExtendedGrammar.format_number)
        super().add_function("$formatNumber", ExtendedGrammar.format_number)
        super().add_transform("formatNumber", ExtendedGrammar.format_number)
        super().add_function("formatBase", ExtendedGrammar.format_base)
        super().add_function("$formatBase", ExtendedGrammar.format_base)
        super().add_transform("formatBase", ExtendedGrammar.format_base)
        super().add_function("formatInteger", ExtendedGrammar.format_integer)
        super().add_function("$formatInteger", ExtendedGrammar.format_integer)
        super().add_transform("formatInteger", ExtendedGrammar.format_integer)
        super().add_function("sum", ExtendedGrammar.sum)
        super().add_function("$sum", ExtendedGrammar.sum)
        super().add_transform("sum", ExtendedGrammar.sum)
        super().add_function("max", ExtendedGrammar.max)
        super().add_function("$max", ExtendedGrammar.max)
        super().add_transform("max", ExtendedGrammar.max)
        super().add_function("min", ExtendedGrammar.min)
        super().add_function("$min", ExtendedGrammar.min)
        super().add_transform("min", ExtendedGrammar.min)
        super().add_function("avg", ExtendedGrammar.avg)
        super().add_function("$avg", ExtendedGrammar.avg)
        super().add_function("average", ExtendedGrammar.avg)
        super().add_function("$average", ExtendedGrammar.avg)
        super().add_transform("avg", ExtendedGrammar.avg)
        super().add_transform("average", ExtendedGrammar.avg)

        """ Boolean functions """

        super().add_function("boolean", ExtendedGrammar.to_boolean)
        super().add_function("$boolean", ExtendedGrammar.to_boolean)
        super().add_transform("boolean", ExtendedGrammar.to_boolean)
        super().add_function("bool", ExtendedGrammar.to_boolean)
        super().add_function("$bool", ExtendedGrammar.to_boolean)
        super().add_transform("bool", ExtendedGrammar.to_boolean)
        super().add_transform("toBoolean", ExtendedGrammar.to_boolean)
        super().add_transform("toBool", ExtendedGrammar.to_boolean)
        super().add_function("not", ExtendedGrammar.not_)
        super().add_function("$not", ExtendedGrammar.not_)
        super().add_transform("not", ExtendedGrammar.not_)

        """ Array functions """

        super().add_function("append", ExtendedGrammar.array_append)
        super().add_function("$append", ExtendedGrammar.array_append)
        super().add_transform("append", ExtendedGrammar.array_append)
        super().add_function("concat", ExtendedGrammar.array_append)
        super().add_function("$concat", ExtendedGrammar.array_append)
        super().add_transform("concat", ExtendedGrammar.array_append)
        super().add_function("reverse", ExtendedGrammar.array_reverse)
        super().add_function("$reverse", ExtendedGrammar.array_reverse)
        super().add_transform("reverse", ExtendedGrammar.array_reverse)
        super().add_function("shuffle", ExtendedGrammar.array_shuffle)
        super().add_function("$shuffle", ExtendedGrammar.array_shuffle)
        super().add_transform("shuffle", ExtendedGrammar.array_shuffle)
        super().add_function("sort", ExtendedGrammar.array_sort)
        super().add_function("$sort", ExtendedGrammar.array_sort)
        super().add_transform("sort", ExtendedGrammar.array_sort)
        super().add_function("order", ExtendedGrammar.array_sort)
        super().add_function("$order", ExtendedGrammar.array_sort)
        super().add_transform("order", ExtendedGrammar.array_sort)
        super().add_function("distinct", ExtendedGrammar.array_distinct)
        super().add_function("$distinct", ExtendedGrammar.array_distinct)
        super().add_transform("distinct", ExtendedGrammar.array_distinct)
        super().add_function("toObject", ExtendedGrammar.array_to_object)
        super().add_function("$toObject", ExtendedGrammar.array_to_object)
        super().add_transform("toObject", ExtendedGrammar.array_to_object)
        super().add_function("fromEntries", ExtendedGrammar.array_to_object)
        super().add_function("$fromEntries", ExtendedGrammar.array_to_object)
        super().add_transform("fromEntries", ExtendedGrammar.array_to_object)
        super().add_function("mapField", ExtendedGrammar.array_mapfield)
        super().add_function("$mapField", ExtendedGrammar.array_mapfield)
        super().add_transform("mapField", ExtendedGrammar.array_mapfield)
        super().add_function("map", self._extended_grammar.array_map)
        super().add_function("$map", self._extended_grammar.array_map)
        super().add_transform("map", self._extended_grammar.array_map)
        super().add_function("any", self._extended_grammar.array_any)
        super().add_function("$any", self._extended_grammar.array_any)
        super().add_transform("any", self._extended_grammar.array_any)
        super().add_function("some", self._extended_grammar.array_any)
        super().add_function("$some", self._extended_grammar.array_any)
        super().add_transform("some", self._extended_grammar.array_any)
        super().add_function("all", self._extended_grammar.array_every)
        super().add_function("$all", self._extended_grammar.array_every)
        super().add_transform("all", self._extended_grammar.array_every)
        super().add_function("every", self._extended_grammar.array_every)
        super().add_function("$every", self._extended_grammar.array_every)
        super().add_transform("every", self._extended_grammar.array_every)
        super().add_function("filter", self._extended_grammar.array_filter)
        super().add_function("$filter", self._extended_grammar.array_filter)
        super().add_transform("filter", self._extended_grammar.array_filter)
        super().add_function("find", self._extended_grammar.array_find)
        super().add_function("$find", self._extended_grammar.array_find)
        super().add_transform("find", self._extended_grammar.array_find)
        super().add_function("reduce", self._extended_grammar.array_reduce)
        super().add_function("$reduce", self._extended_grammar.array_reduce)
        super().add_transform("reduce", self._extended_grammar.array_reduce)

        """ Object functions """

        super().add_function("keys", ExtendedGrammar.object_keys)
        super().add_function("$keys", ExtendedGrammar.object_keys)
        super().add_transform("keys", ExtendedGrammar.object_keys)
        super().add_function("values", ExtendedGrammar.object_values)
        super().add_function("$values", ExtendedGrammar.object_values)
        super().add_transform("values", ExtendedGrammar.object_values)
        super().add_function("entries", ExtendedGrammar.object_entries)
        super().add_function("$entries", ExtendedGrammar.object_entries)
        super().add_transform("entries", ExtendedGrammar.object_entries)
        super().add_function("merge", ExtendedGrammar.object_merge)
        super().add_function("$merge", ExtendedGrammar.object_merge)
        super().add_transform("merge", ExtendedGrammar.object_merge)

        """ Date functions """
        super().add_function("now", ExtendedGrammar.now)
        super().add_function("$now", ExtendedGrammar.now)
        super().add_function("millis", ExtendedGrammar.millis)
        super().add_function("$millis", ExtendedGrammar.millis)
        super().add_function("millisToDateTime", ExtendedGrammar.to_datetime)
        super().add_function("$millisToDateTime", ExtendedGrammar.to_datetime)
        super().add_transform("millisToDateTime", ExtendedGrammar.to_datetime)
        super().add_function("fromMillis", ExtendedGrammar.to_datetime)
        super().add_function("$fromMillis", ExtendedGrammar.to_datetime)
        super().add_transform("fromMillis", ExtendedGrammar.to_datetime)
        super().add_function("toDateTime", ExtendedGrammar.to_datetime)
        super().add_function("$toDateTime", ExtendedGrammar.to_datetime)
        super().add_transform("toDateTime", ExtendedGrammar.to_datetime)
        super().add_function("dateTimeString", ExtendedGrammar.to_datetime)
        super().add_function("dateTimeToMillis", ExtendedGrammar.to_millis)
        super().add_function("$dateTimeToMillis", ExtendedGrammar.to_millis)
        super().add_transform("dateTimeToMillis", ExtendedGrammar.to_millis)
        super().add_function("toMillis", ExtendedGrammar.to_millis)
        super().add_function("$toMillis", ExtendedGrammar.to_millis)
        super().add_transform("toMillis", ExtendedGrammar.to_millis)
        super().add_function("dateTimeAdd", ExtendedGrammar.datetime_add)
        super().add_function("$dateTimeAdd", ExtendedGrammar.datetime_add)
        super().add_transform("dateTimeAdd", ExtendedGrammar.datetime_add)

        """ Misc """

        super().add_function("eval", ExtendedGrammar._eval)
        super().add_function("$eval", ExtendedGrammar._eval)
        super().add_transform("eval", ExtendedGrammar._eval)
        super().add_function("uuid", ExtendedGrammar.uuid)
        super().add_function("$uuid", ExtendedGrammar.uuid)
        super().add_function("uid", ExtendedGrammar.uuid)



from pyjexl.parser import (
    ArrayLiteral,
    BinaryExpression,
    ConditionalExpression,
    Function,
    Identifier,
    Literal,
    ObjectLiteral,
    Transform,
    UnaryExpression,
    FilterExpression,
)

from . import default_config, DefaultParser


_ops = {}
_ops.update(default_config.binary_operators)
_ops.update(default_config.unary_operators)


def test_literal():
    assert DefaultParser().parse("1") == Literal(1.0)


def test_binary_expression():
    assert DefaultParser().parse("1+2") == BinaryExpression(
        operator=_ops["+"], left=Literal(1), right=Literal(2)
    )


def test_binary_expression_priority_right():
    assert DefaultParser().parse("2+3*4") == BinaryExpression(
        operator=_ops["+"],
        left=Literal(2),
        right=BinaryExpression(
            operator=_ops["*"],
            left=Literal(3),
            right=Literal(4),
        ),
    )


def test_binary_expression_priority_left():
    assert DefaultParser().parse("2*3+4") == BinaryExpression(
        operator=_ops["+"],
        left=BinaryExpression(
            operator=_ops["*"],
            left=Literal(2),
            right=Literal(3),
        ),
        right=Literal(4),
    )


def test_binary_expression_encapsulation():
    assert DefaultParser().parse("2+3*4==5/6-7") == BinaryExpression(
        operator=_ops["=="],
        left=BinaryExpression(
            operator=_ops["+"],
            left=Literal(2),
            right=BinaryExpression(
                operator=_ops["*"], left=Literal(3), right=Literal(4)
            ),
        ),
        right=BinaryExpression(
            operator=_ops["-"],
            left=BinaryExpression(
                operator=_ops["/"], left=Literal(5), right=Literal(6)
            ),
            right=Literal(7),
        ),
    )


def test_unary_operator():
    assert DefaultParser().parse("1*!!true-2") == BinaryExpression(
        operator=_ops["-"],
        left=BinaryExpression(
            operator=_ops["*"],
            left=Literal(1),
            right=UnaryExpression(
                operator=_ops["!"],
                right=UnaryExpression(operator=_ops["!"], right=Literal(True)),
            ),
        ),
        right=Literal(2),
    )


def test_subexpression():
    assert DefaultParser().parse("(2+3)*4") == BinaryExpression(
        operator=_ops["*"],
        left=BinaryExpression(operator=_ops["+"], left=Literal(2), right=Literal(3)),
        right=Literal(4),
    )


def test_nested_subexpression():
    assert DefaultParser().parse("(4*(2+3))/5") == BinaryExpression(
        operator=_ops["/"],
        left=BinaryExpression(
            operator=_ops["*"],
            left=Literal(4),
            right=BinaryExpression(
                operator=_ops["+"], left=Literal(2), right=Literal(3)
            ),
        ),
        right=Literal(5),
    )


def test_object_literal():
    assert DefaultParser().parse('{foo: "bar", tek: 1+2}') == ObjectLiteral(
        {
            "foo": Literal("bar"),
            "tek": BinaryExpression(
                operator=_ops["+"], left=Literal(1), right=Literal(2)
            ),
        }
    )


def test_nested_object_literals():
    assert DefaultParser().parse('{foo: {bar: "tek"}}') == ObjectLiteral(
        {"foo": ObjectLiteral({"bar": Literal("tek")})}
    )


def test_empty_object_literals():
    assert DefaultParser().parse("{}") == ObjectLiteral({})


def test_array_literals():
    assert DefaultParser().parse('["foo", 1+2]') == ArrayLiteral(
        [
            Literal("foo"),
            BinaryExpression(operator=_ops["+"], left=Literal(1), right=Literal(2)),
        ]
    )


def test_nexted_array_literals():
    assert DefaultParser().parse('["foo", ["bar", "tek"]]') == ArrayLiteral(
        [Literal("foo"), ArrayLiteral([Literal("bar"), Literal("tek")])]
    )


def test_empty_array_literals():
    assert DefaultParser().parse("[]") == ArrayLiteral([])


def test_chained_identifiers():
    assert DefaultParser().parse("foo.bar.baz + 1") == BinaryExpression(
        operator=_ops["+"],
        left=Identifier("baz", subject=Identifier("bar", subject=Identifier("foo"))),
        right=Literal(1),
    )


def test_function():
    assert DefaultParser().parse('fn1("bar")') == Function(
        name="fn1", args=[Literal("bar")]
    )


def test_function_no_args():
    assert DefaultParser().parse("fn2()") == Function(name="fn2", args=[])


def test_functions_object_arg():
    assert DefaultParser().parse('fn1({bar:"tek"})') == Function(
        name="fn1", args=[ObjectLiteral({"bar": Literal("tek")})]
    )


def test_functions_multiple_arguments():
    assert DefaultParser().parse('fn1("bar","baz")') == Function(
        name="fn1", args=[Literal("bar"), Literal("baz")]
    )


def test_transforms():
    assert DefaultParser().parse('foo|tr1|tr2.baz|tr3({bar:"tek"})') == Transform(
        name="tr3",
        args=[ObjectLiteral({"bar": Literal("tek")})],
        subject=Identifier(
            "baz",
            subject=Transform(
                name="tr2",
                args=[],
                subject=Transform(name="tr1", args=[], subject=Identifier("foo")),
            ),
        ),
    )


def test_transforms_multiple_arguments():
    assert DefaultParser().parse('foo|bar("tek", 5, true)') == Transform(
        name="bar",
        args=[Literal("tek"), Literal(5), Literal(True)],
        subject=Identifier("foo"),
    )


def test_filters():
    assert DefaultParser().parse('foo[1][.bar[0]=="tek"].baz') == Identifier(
        value="baz",
        subject=FilterExpression(
            relative=True,
            expression=BinaryExpression(
                operator=_ops["=="],
                left=FilterExpression(
                    relative=False,
                    expression=Literal(0),
                    subject=Identifier("bar", relative=True),
                ),
                right=Literal("tek"),
            ),
            subject=FilterExpression(
                relative=False, expression=Literal(1), subject=Identifier("foo")
            ),
        ),
    )


def test_attribute_all_operands():
    assert DefaultParser().parse('"foo".length + {foo: "bar"}.foo') == BinaryExpression(
        operator=_ops["+"],
        left=Identifier("length", subject=Literal("foo")),
        right=Identifier(value="foo", subject=ObjectLiteral({"foo": Literal("bar")})),
    )


def test_attribute_subexpression():
    assert DefaultParser().parse('("foo" + "bar").length') == Identifier(
        value="length",
        subject=BinaryExpression(
            operator=_ops["+"], left=Literal("foo"), right=Literal("bar")
        ),
    )


def test_attribute_array():
    assert DefaultParser().parse('["foo", "bar"].length') == Identifier(
        value="length", subject=ArrayLiteral([Literal("foo"), Literal("bar")])
    )


def test_identifier_filter_expression():
    assert DefaultParser().parse('foo.bar["baz"]') == FilterExpression(
        expression=Literal("baz"),
        subject=Identifier(value="bar", subject=Identifier(value="foo")),
    )


def test_ternary_expression():
    assert DefaultParser().parse("foo ? 1 : 0") == ConditionalExpression(
        test=Identifier("foo"), consequent=Literal(1), alternate=Literal(0)
    )


def test_nested_grouped_ternary_expression():
    assert DefaultParser().parse("foo ? (bar ? 1 : 2) : 3") == ConditionalExpression(
        test=Identifier("foo"),
        consequent=ConditionalExpression(
            test=Identifier("bar"), consequent=Literal(1), alternate=Literal(2)
        ),
        alternate=Literal(3),
    )


def test_nested_non_grouped_ternary_expression():
    assert DefaultParser().parse("foo ? bar ? 1 : 2 : 3") == ConditionalExpression(
        test=Identifier("foo"),
        consequent=ConditionalExpression(
            test=Identifier("bar"), consequent=Literal(1), alternate=Literal(2)
        ),
        alternate=Literal(3),
    )


def test_object_ternary_expression():
    assert DefaultParser().parse('foo ? {bar: "tek"} : "baz"') == ConditionalExpression(
        test=Identifier("foo"),
        consequent=ObjectLiteral({"bar": Literal("tek")}),
        alternate=Literal("baz"),
    )


def test_functions_in_ternary_expression():
    assert DefaultParser().parse('foo ? bar(1) : baz("tek")') == ConditionalExpression(
        test=Identifier("foo"),
        consequent=Function(name="bar", args=[Literal(1)]),
        alternate=Function(name="baz", args=[Literal("tek")]),
    )


def test_complex_binary_operator_balancing():
    assert DefaultParser().parse("a.b == c.d") == BinaryExpression(
        operator=_ops["=="],
        left=Identifier("b", subject=Identifier("a")),
        right=Identifier("d", subject=Identifier("c")),
    )


def test_arbitrary_whitespace():
    assert DefaultParser().parse("\t2\r\n+\n\r3\n\n") == BinaryExpression(
        operator=_ops["+"], left=Literal(2), right=Literal(3)
    )



import unittest
from pyjexl import JexlExtended


class JexlExtendedTests(unittest.TestCase):

    def setUp(self) -> None:
        self.jexl = JexlExtended()

    def test_string(self):
        self.assertEqual(self.jexl.evaluate("string(123)"), "123")
        self.assertEqual(self.jexl.evaluate("123456|string"), "123456")
        self.assertEqual(
            self.jexl.evaluate("""{a:123456}|string"""), """{"a":123456}"""
        )

    def test_length(self):
        self.assertEqual(self.jexl.evaluate("'test123'|length"), 7)
        self.assertEqual(self.jexl.evaluate('["a",1,"b"]|length'), 3)
        self.assertEqual(self.jexl.evaluate('$length(["a",1,"b"])'), 3)
        self.assertEqual(self.jexl.evaluate("""{a:1,b:2,c:3}|length"""), 3)

    def test_substring(self):
        self.assertEqual(self.jexl.evaluate("substring(123456,2,2)"), "34")
        self.assertEqual(self.jexl.evaluate("$substring('test',(-2))"), "st")
        self.assertEqual(
            self.jexl.evaluate("$substring($string({a:123456}, true),0,1)"), "{"
        )

    def test_substring_before(self):
        self.assertEqual(
            self.jexl.evaluate('"hello world"|substringBefore(" ")'), "hello"
        )
        self.assertEqual(
            self.jexl.evaluate('substringBefore("hello world", "o")'), "hell"
        )
        self.assertEqual(
            self.jexl.evaluate('substringBefore("hello world", "x")'), "hello world"
        )
        self.assertEqual(self.jexl.evaluate("substringBefore(123456,2)"), "1")

    def test_substring_after(self):
        self.assertEqual(
            self.jexl.evaluate('"hello world"|substringAfter(" ")'), "world"
        )
        self.assertEqual(
            self.jexl.evaluate('substringAfter("hello world", "o")'), " world"
        )
        self.assertEqual(self.jexl.evaluate('substringAfter("hello world", "x")'), "")
        self.assertEqual(self.jexl.evaluate("substringAfter(123456,2)"), "3456")

    def test_uppercase(self):
        self.assertEqual(self.jexl.evaluate('uppercase("hello world")'), "HELLO WORLD")
        self.assertEqual(self.jexl.evaluate("uppercase(123456)"), "123456")
        self.assertEqual(self.jexl.evaluate("'baz'|uppercase"), "BAZ")

    def test_lowercase(self):
        self.assertEqual(self.jexl.evaluate('lowercase("HELLO WORLD")'), "hello world")
        self.assertEqual(self.jexl.evaluate("lowercase(123456)"), "123456")
        self.assertEqual(self.jexl.evaluate('$lowercase("FOObar")'), "foobar")
        self.assertEqual(self.jexl.evaluate('"FOObar"|lower'), "foobar")

    def test_camel_pascal_case(self):
        self.assertEqual(self.jexl.evaluate("'foo bar '|camelCase"), "fooBar")
        self.assertEqual(self.jexl.evaluate('$camelCase("Foo_bar")'), "fooBar")
        self.assertEqual(self.jexl.evaluate("'FooBar'|toCamelCase"), "fooBar")
        self.assertEqual(self.jexl.evaluate("'Foo-bar'|toCamelCase"), "fooBar")
        self.assertEqual(self.jexl.evaluate("'foo bar'|toPascalCase"), "FooBar")
        self.assertEqual(self.jexl.evaluate("'fooBar'|toPascalCase"), "FooBar")
        self.assertEqual(self.jexl.evaluate("'Foo_bar'|toPascalCase"), "FooBar")

    def test_trim_pad(self):
        self.assertEqual(self.jexl.evaluate('trim(" baz  ")'), "baz")
        self.assertEqual(self.jexl.evaluate('trim("  baz  ")'), "baz")
        self.assertEqual(self.jexl.evaluate('trim("__baz--","--")'), "__baz")
        self.assertEqual(self.jexl.evaluate('pad("foo",5)'), "foo  ")
        self.assertEqual(self.jexl.evaluate('pad("foo",-5,0)'), "00foo")

    def test_contains(self):
        self.assertTrue(self.jexl.evaluate("'foo-bar'|contains('bar')"))
        self.assertFalse(self.jexl.evaluate("'foo-bar'|contains('baz')"))
        self.assertFalse(self.jexl.evaluate('["foo-bar"]|contains("bar")'))
        self.assertTrue(self.jexl.evaluate('["foo-bar"]|contains("foo-bar")'))
        self.assertTrue(self.jexl.evaluate('["baz", "foo", "bar"]|contains("bar")'))

    def test_split(self):
        self.assertEqual(self.jexl.evaluate('split("foo-bar", "-")'), ["foo", "bar"])
        self.assertEqual(self.jexl.evaluate('split("foo-bar", "-")[1]'), "bar")
        self.assertEqual(self.jexl.evaluate('split("foo-bar", "-")[0]'), "foo")
        self.assertEqual(
            self.jexl.evaluate('split("foo-bar", "-")[0]|uppercase'), "FOO"
        )
        self.assertEqual(
            self.jexl.evaluate('split("foo-bar", "-")[1]|lowercase'), "bar"
        )
        self.assertEqual(
            self.jexl.evaluate('split("foo-bar", "-")[1]|substring(0,1)'), "b"
        )

    def test_join(self):
        self.assertEqual(self.jexl.evaluate('join(["foo", "bar"], "-")'), "foo-bar")
        self.assertEqual(self.jexl.evaluate('join(["foo", "bar"], "")'), "foobar")
        self.assertEqual(self.jexl.evaluate('["foo", "bar"]|join'), "foo,bar")
        # self.assertEqual(
        #     self.jexl.evaluate('"f,b,a,d,e,c"|split(",")|sort|join'), "a,b,c,d,e,f"
        # )
        # self.assertEqual(
        #     self.jexl.evaluate('"f,b,a,d,e,c"|split(",")|sort|join("")'), "abcdef"
        # )

    def test_replace(self):
        self.assertEqual(self.jexl.evaluate('replace("foo-bar", "-", "_")'), "foo_bar")
        self.assertEqual(self.jexl.evaluate('replace("foo-bar---", "-", "")'), "foobar")
        self.assertEqual(
            self.jexl.evaluate('"123ab123ab123ab"|replace("123")'), "ababab"
        )

    def test_base64(self):
        self.assertEqual(self.jexl.evaluate('base64Encode("foobar")'), "Zm9vYmFy")
        self.assertEqual(self.jexl.evaluate('base64Decode("Zm9vYmFy")'), "foobar")

    def test_form_url_encoded(self):
        self.assertEqual(
            self.jexl.evaluate('{foo:"bar",baz:"tek"}|formUrlEncoded'),
            "foo=bar&baz=tek",
        )

    def test_number(self):
        self.assertEqual(self.jexl.evaluate('$number("1")'), 1)
        self.assertEqual(self.jexl.evaluate('$number("1.1")'), 1.1)
        self.assertEqual(self.jexl.evaluate('$number("-1.1")'), -1.1)
        self.assertEqual(self.jexl.evaluate("$number(-1.1)"), -1.1)
        self.assertEqual(self.jexl.evaluate("$number(-1.1)|floor"), -2)
        self.assertEqual(self.jexl.evaluate('$number("10.6")|ceil'), 11)
        self.assertEqual(self.jexl.evaluate("10.123456|round(2)"), 10.12)
        self.assertEqual(self.jexl.evaluate("10.123456|toInt"), 10)
        self.assertEqual(self.jexl.evaluate('"10.123456"|toInt'), 10)
        self.assertEqual(self.jexl.evaluate("'16325'|toInt"), 16325)
        self.assertEqual(self.jexl.evaluate("(9/2)|toInt"), 4)
        self.assertEqual(self.jexl.evaluate("3|power(2)"), 9)
        self.assertEqual(self.jexl.evaluate("3|power"), 9)
        self.assertEqual(self.jexl.evaluate("9|sqrt"), 3)
        self.assertEqual(self.jexl.evaluate("random() < 1"), True)

    def test_formatting(self):
        self.assertEqual(
            self.jexl.evaluate('16325.62|formatNumber("0,0.000")'), "16,325.620"
        )
        self.assertEqual(
            self.jexl.evaluate('16325.62|formatNumber("0.000")'), "16325.620"
        )
        self.assertEqual(self.jexl.evaluate("12|formatBase(16)"), "c")
        self.assertEqual(
            self.jexl.evaluate('16325.62|formatInteger("0000000")'), "0016325"
        )

    def test_numeric_aggregations(self):
        self.assertEqual(self.jexl.evaluate("[1,2,3]|sum"), 6)
        self.assertEqual(self.jexl.evaluate("sum(1,2,3,4,5)"), 15)
        self.assertEqual(self.jexl.evaluate("[1,3]|sum(1,2,3,4,5)"), 19)
        self.assertEqual(self.jexl.evaluate("[1,3]|sum([1,2,3,4,5])"), 19)
        self.assertEqual(self.jexl.evaluate("[1,3]|max([1,2,3,4,5])"), 5)
        self.assertEqual(self.jexl.evaluate("[2,3]|min([1,2,3,4,5])"), 1)
        self.assertEqual(self.jexl.evaluate("[2,3]|min(1,2,3,4,5)"), 1)
        self.assertEqual(self.jexl.evaluate("[4,5,6]|avg"), 5)

    def test_booleans(self):
        self.assertTrue(self.jexl.evaluate("1|toBoolean"))
        self.assertTrue(self.jexl.evaluate("3|toBoolean"))
        self.assertTrue(self.jexl.evaluate("'1'|toBoolean"))
        self.assertIsNone(self.jexl.evaluate("'2'|toBoolean"))
        self.assertIsNone(self.jexl.evaluate("'a'|toBool"))
        self.assertIsNone(self.jexl.evaluate("''|toBool"))
        self.assertFalse(self.jexl.evaluate("0|toBool"))
        self.assertFalse(self.jexl.evaluate("0.0|toBool"))
        self.assertFalse(self.jexl.evaluate("'false'|toBool"))
        self.assertFalse(self.jexl.evaluate("'False'|toBool"))
        self.assertFalse(self.jexl.evaluate("'fALSE'|toBool"))
        self.assertTrue(self.jexl.evaluate("'tRUE       '|toBoolean"))
        self.assertTrue(self.jexl.evaluate("'False'|toBool|not"))
        self.assertFalse(self.jexl.evaluate("'TRUE'|toBool|not"))

    def test_arrays(self):
        self.assertEqual(
            self.jexl.evaluate('["foo", "bar", "baz"]|append("tek")'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate('["foo", "bar"]|append(["baz","tek"])'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate('"foo"|append(["bar", "baz","tek"])'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate('"foo"|append("bar", "baz","tek")'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate('["tek", "baz", "bar", "foo"]|reverse'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate('["tek", "baz", "bar", "foo", "foo"]|reverse|distinct'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate("{foo:0, bar:1, baz:2, tek:3}|keys"),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate('{a:"foo", b:"bar", c:"baz", d:"tek"}|values'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate(
                '[{name:"foo"}, {name:"bar"}, {name:"baz"}, {name:"tek"}]|mapField("name")'
            ),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate(
                '[{name:"tek",age:32}, {name:"bar",age:34}, {name:"baz",age:33}, {name:"foo",age:35}]|sort("age",true)|mapField("name")'
            ),
            ["tek", "baz", "bar", "foo"],
        )
        self.assertEqual(
            self.jexl.evaluate('["foo"]|append(["tek","baz","bar"]|sort)'),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate(
                '["foo"]|append(["tek", "baz", "bar", "foo", "foo"]|filter("value != \'foo\'")|sort)'
            ),
            ["foo", "bar", "baz", "tek"],
        )

    def test_map(self):
        context = {
            "assoc": [
                {"lastName": "Archer", "age": 32},
                {"lastName": "Poovey", "age": 34},
                {"lastName": "Figgis", "age": 45},
            ]
        }
        self.assertEqual(
            self.jexl.evaluate(
                '[{name:"foo"}, {name:"bar"}, {name:"baz"}, {name:"tek"}]|map("value.name")'
            ),
            ["foo", "bar", "baz", "tek"],
        )
        self.assertEqual(
            self.jexl.evaluate(
                '[{name:"tek",age:32}, {name:"bar",age:34}, {name:"baz",age:33}, {name:"foo",age:35}]|map("value.age")'
            ),
            [32, 34, 33, 35],
        )
        self.assertEqual(
            self.jexl.evaluate("assoc|map('value.age')", context),
            [32, 34, 45],
        )
        self.assertEqual(
            self.jexl.evaluate("assoc|map('value.lastName')", context),
            ["Archer", "Poovey", "Figgis"],
        )
        self.assertEqual(
            self.jexl.evaluate("assoc|map('value.age + index')", context),
            [32, 35, 47],
        )
        self.assertEqual(
            self.jexl.evaluate(
                "assoc|map('value.age + array[.age <= value.age][0].age + index')",
                context,
            ),
            [64, 67, 79],
        )
        self.assertEqual(
            self.jexl.evaluate("assoc|map('value.age')|avg", context),
            37,
        )

    def test_any_all(self):
        context = {
            "assoc": [
                {"lastName": "Archer", "age": 32},
                {"lastName": "Poovey", "age": 34},
                {"lastName": "Figgis", "age": 45},
            ]
        }
        self.assertTrue(
            self.jexl.evaluate(
                '[{name:"foo"}, {name:"bar"}, {name:"baz"}, {name:"tek"}]|any("value.name==\'foo\'")'
            )
        )
        self.assertTrue(self.jexl.evaluate("assoc|every('value.age>30')", context))
        self.assertFalse(self.jexl.evaluate("assoc|every('value.age>40')", context))
        self.assertTrue(self.jexl.evaluate("assoc|some('value.age>40')", context))
        self.assertTrue(
            self.jexl.evaluate("assoc|some('value.lastName=='Figgis'')", context)
        )
        self.assertTrue(
            self.jexl.evaluate("assoc|map('value.age')|some('value>30')", context)
        )

    def test_reduce(self):
        context = {
            "assoc": [
                {"lastName": "Archer", "age": 32},
                {"lastName": "Poovey", "age": 34},
                {"lastName": "Figgis", "age": 45},
            ]
        }
        self.assertEqual(
            self.jexl.evaluate("assoc|reduce('accumulator + value.age', 0)", context),
            111,
        )
        self.assertEqual(
            self.jexl.evaluate(
                "assoc|reduce('(value.age > array|map(\"value.age\")|avg) ? accumulator|append(value.age) : accumulator', [])",
                context,
            ),
            [45],
        )
        self.assertEqual(
            self.jexl.evaluate(
                "assoc|reduce('(value.age < array|map(\"value.age\")|avg) ? accumulator|append(value.age) : accumulator', [])[1]",
                context,
            ),
            34,
        )

    def test_objects(self):
        self.assertEqual(
            self.jexl.evaluate("$merge({foo:'bar'},{baz:'tek'})"),
            {"foo": "bar", "baz": "tek"},
        )
        self.assertEqual(
            self.jexl.evaluate("{foo:'bar'}|merge({baz:'tek'})"),
            {"foo": "bar", "baz": "tek"},
        )
        self.assertEqual(
            self.jexl.evaluate("[{foo:'bar'},{baz:'tek'}]|merge"),
            {"foo": "bar", "baz": "tek"},
        )
        self.assertEqual(
            self.jexl.evaluate("[{foo:'bar'}]|merge([{baz:'tek'}])"),
            {"foo": "bar", "baz": "tek"},
        )
        self.assertEqual(
            self.jexl.evaluate('[["foo","bar"],["baz","tek"]]|toObject'),
            {"foo": "bar", "baz": "tek"},
        )
        self.assertEqual(
            self.jexl.evaluate('["foo","bar"]|toObject(true)'),
            {"foo": True, "bar": True},
        )
        self.assertEqual(
            self.jexl.evaluate('["a","b","c"]|toObject(true)'),
            {"a": True, "b": True, "c": True},
        )

    def test_convert_time_zone(self):
        # Windows timezone
        self.assertEqual(
            self.jexl.evaluate(
                "'2025-11-26T12:00:00Z'|convertTimeZone('Pacific Standard Time')"
            ),
            "2025-11-26T04:00:00.0000000-08:00",
        )
        self.assertEqual(
            self.jexl.evaluate(
                "'2025-06-26T12:00:00Z'|convertTimeZone('Pacific Standard Time')"
            ),
            "2025-06-26T05:00:00.0000000-07:00",
        )
        # IANA timezone
        self.assertEqual(
            self.jexl.evaluate(
                "'2025-06-26T12:00:00Z'|convertTimeZone('Europe/Amsterdam')"
            ),
            "2025-06-26T14:00:00.0000000+02:00",
        )
        self.assertEqual(
            self.jexl.evaluate("'2025-11-26T12:00:00Z'|convertTimeZone('UTC')"),
            "2025-11-26T12:00:00.0000000+00:00",
        )
        # Fixed offsets
        self.assertEqual(
            self.jexl.evaluate("'2025-11-26T12:00:00Z'|convertTimeZone('+02:00')"),
            "2025-11-26T14:00:00.0000000+02:00",
        )
        self.assertEqual(
            self.jexl.evaluate("'2025-06-26T12:00:00Z'|convertTimeZone('-08:00')"),
            "2025-06-26T04:00:00.0000000-08:00",
        )
        self.assertEqual(
            self.jexl.evaluate("'2025-12-26T12:00:00Z'|convertTimeZone('-08:00')"),
            "2025-12-26T04:00:00.0000000-08:00",
        )



class JEXLException(Exception):
    """Base class for all pyjexl exceptions."""


class ParseError(JEXLException):
    """An error during the parsing of an expression."""


class InvalidOperatorError(ParseError):
    """An invalid operator was used."""


class EvaluationError(JEXLException):
    """An error during evaluation of a parsed expression."""


class MissingTransformError(EvaluationError):
    """An unregistered transform was used."""


class MissingFunctionError(EvaluationError):
    """An unregistered function was used."""
